
from bson.py3compat import string_type

from . import errors
from .collection import MontyCollection
from .base import BaseObject
from .engine.helpers import encode_


class MontyDatabase(BaseObject):

    def __init__(self, client, name, codec_options=None, write_concern=None):
        """
        """
        super(MontyDatabase, self).__init__(
            codec_options or client.codec_options,
            write_concern or client.write_concern)

        self._client = client
        self._name = name
        self._components = (self,)

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return (self._client == other.client and
                    self._name == other.name)
        return NotImplemented

    def __ne__(self, other):
        return not self == other

    def __repr__(self):
        return "MontyDatabase({!r}, {!r})".format(self._client, self._name)

    def __getattr__(self, name):
        if name.startswith('_'):
            raise AttributeError(
                "MontyDatabase has no attribute {0!r}. To access the {0}"
                " collection, use database[{0!r}].".format(name))
        return self.get_collection(name)

    def __getitem__(self, key):
        return self.get_collection(key)

    @property
    def name(self):
        """
        The name of this Database.
        """
        return self._name

    @property
    def client(self):
        """
        """
        return self._client

    def collection_names(self):
        return self.client._storage.collection_list(self)

    def create_collection(self,
                          name,
                          codec_options=None,
                          write_concern=None,
                          **kwargs):
        """
        Create a collection, before any insertion,
        raise error if exists.
        """
        if self.client._storage.collection_exists(self, name):
            error_msg = "collection {} already exists".format(encode_(name))
            raise errors.CollectionInvalid(error_msg)
        else:
            collection = self.get_collection(name,
                                             codec_options,
                                             write_concern,
                                             **kwargs)
            self.client._storage.collection_create(self, name)
            return collection

    def drop_collection(self, name_or_collection):
        """
        """
        name = name_or_collection
        if isinstance(name_or_collection, MontyCollection):
            name = name_or_collection.name
        elif not isinstance(name_or_collection, string_type):
            raise TypeError("name_or_collection must be an instance of "
                            "basestring")
        self.client._storage.collection_drop(self, name)

    def get_collection(self,
                       name,
                       codec_options=None,
                       write_concern=None,
                       **kwargs):
        """
        Get a collection.
        """
        # verify collection name
        is_invaild = False
        for invaild_char in ('$', '\0', '\x00'):
            if invaild_char in name:
                is_invaild = True
        if name.startswith('system.'):
            is_invaild = True

        if is_invaild or not name:
            raise errors.OperationFailure("Invaild collection name.")
        else:
            return MontyCollection(self,
                                   name,
                                   False,
                                   codec_options,
                                   write_concern,
                                   **kwargs)



import warnings
import collections
from copy import deepcopy
from bson import ObjectId
from bson.py3compat import string_type

from .base import (
    BaseObject,
    validate_is_mapping,
    validate_ok_for_update,
    validate_ok_for_replace,
    validate_list_or_none,
    validate_boolean,
    command_coder,
)

from .cursor import MontyCursor
from .engine.core import FieldWalker, Weighted
from .engine.queries import QueryFilter
from .engine.update import Updator
from .engine.helpers import is_duckument_type, Counter

from .storage.errors import StorageDuplicateKeyError
from .errors import (
    DuplicateKeyError,
    BulkWriteError,
)

from .results import (BulkWriteResult,
                      DeleteResult,
                      InsertOneResult,
                      InsertManyResult,
                      UpdateResult)


class MontyCollection(BaseObject):

    def __init__(self, database, name, create=False,
                 codec_options=None, write_concern=None, **kwargs):
        """
        """
        super(MontyCollection, self).__init__(
            codec_options or database.codec_options,
            write_concern or database.write_concern)

        self._storage = database.client._storage

        self._database = database
        self._name = name
        self._components = (database, self)

    def __repr__(self):
        return "MontyCollection({!r}, {!r})".format(self._database, self._name)

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return (self._database == other.database and
                    self._name == other.name)
        return NotImplemented

    def __ne__(self, other):
        return not self == other

    def __getattr__(self, name):
        if name.startswith('_'):
            full_name = ".".join((self._name, name))
            raise AttributeError(
                "MontyCollection has no attribute {0!r}. To access the {1}"
                " collection, use database[{1!r}].".format(name, full_name))
        return self.__getitem__(name)

    def __getitem__(self, key):
        return self._database.get_collection(".".join((self._name, key)))

    @property
    def full_name(self):
        """
        """
        return u".".join((self._database.name, self._name))

    @property
    def name(self):
        """
        """
        return self._name

    @property
    def database(self):
        """
        """
        return self._database

    def with_options(self, codec_options=None, write_concern=None):
        raise NotImplementedError("Not implemented.")

    def bulk_write(self, requests, ordered=True):
        raise NotImplementedError("Not implemented.")

    def insert_one(self,
                   document,
                   bypass_document_validation=False, *args, **kwargs):
        """
        """
        if bypass_document_validation:
            pass

        if "_id" not in document:
            document["_id"] = ObjectId()

        try:
            result = self._storage.write_one(self, document)
        except StorageDuplicateKeyError:
            message = ("E11000 duplicate key error collection: %s index: "
                       "_id_ dup key: { : \"%s\" }" % (self.full_name,
                                                       str(document["_id"])))
            details = {"index": 0, "code": 11000, "errmsg": message}
            raise DuplicateKeyError(message, code=11000, details=details)

        return InsertOneResult(result)

    def insert_many(self,
                    documents,
                    ordered=True,
                    bypass_document_validation=False, *args, **kwargs):
        """
        """
        if not isinstance(documents, collections.Iterable) or not documents:
            raise TypeError("documents must be a non-empty list")

        if bypass_document_validation:
            pass

        def set_id(doc):
            if "_id" not in doc:
                doc["_id"] = ObjectId()
            # Keep _id in track for error message
            return doc["_id"]

        counter = Counter(iter(documents), job_on_each=set_id)

        try:
            result = self._storage.write_many(self, counter, ordered)
        except StorageDuplicateKeyError:
            message = ("E11000 duplicate key error collection: %s index: "
                       "_id_ dup key: { : \"%s\" }" % (self.full_name,
                                                       str(counter.data)))
            index = counter.count - 1
            result = {
                "writeErrors": [{"index": index,
                                 "code": 11000,
                                 "errmsg": message,
                                 "op": documents[index]}],
                "writeConcernErrors": [],
                "nInserted": index,
                "nUpserted": 0,
                "nMatched": 0,
                "nModified": 0,
                "nRemoved": 0,
                "upserted": [],
            }
            raise BulkWriteError(result)

        return InsertManyResult(result)

    def replace_one(self,
                    filter,
                    replacement,
                    upsert=False,
                    bypass_document_validation=False, *args, **kwargs):
        """
        """
        validate_is_mapping("filter", filter)
        validate_ok_for_replace(replacement)
        validate_boolean("upsert", upsert)

        filter, = command_coder(filter, codec_op=self._database.codec_options)

        if bypass_document_validation:
            pass

        raw_result = {"n": 0, "nModified": 0}
        # updator = Updator(replacement)
        try:
            fw = next(self._internal_scan_query(filter))
        except StopIteration:
            if upsert:
                if "_id" not in replacement:
                    replacement["_id"] = ObjectId()
                raw_result["upserted"] = replacement["_id"]
                raw_result["n"] = 1
                self._storage.write_one(self, replacement, check_keys=False)
        else:
            raw_result["n"] = 1
            if fw.doc != replacement:
                replacement["_id"] = fw.doc["_id"]
                self._storage.update_one(self, replacement)
                raw_result["nModified"] = 1

        return UpdateResult(raw_result)

    def _internal_scan_query(self, query_spec):
        """An interanl document generator for update"""
        queryfilter = QueryFilter(query_spec)
        documents = self._storage.query(MontyCursor(self), 0)
        first_matched = None
        for doc in documents:
            if queryfilter(doc):
                first_matched = queryfilter.fieldwalker
                break

        if first_matched:
            yield first_matched  # for try statement to test update or insert
            yield first_matched  # start update, yield again
            # continue iter documents(generator)
            for doc in documents:
                if queryfilter(doc):
                    yield queryfilter.fieldwalker

    def _internal_upsert(self, query_spec, updator, raw_result):
        """Internal document upsert"""
        doc_cls = self._database.codec_options.document_class

        def _remove_dollar_key(doc):
            if is_duckument_type(doc):
                new_doc = doc_cls()
                fields = list(doc.keys())
                for field in fields:
                    if field[:1] == "$" or "." in field:
                        continue
                    new_doc[field] = _remove_dollar_key(doc[field])
                return new_doc
            else:
                return doc

        document = _remove_dollar_key(deepcopy(query_spec))
        if "_id" not in document:
            document["_id"] = ObjectId()
        raw_result["upserted"] = document["_id"]
        raw_result["n"] = 1

        fieldwalker = FieldWalker(document)
        updator(fieldwalker, do_insert=True)
        self._storage.write_one(self, fieldwalker.doc)

    def update_one(self,
                   filter,
                   update,
                   upsert=False,
                   bypass_document_validation=False,
                   array_filters=None, *args, **kwargs):
        """
        """
        validate_is_mapping("filter", filter)
        validate_ok_for_update(update)
        validate_list_or_none('array_filters', array_filters)
        validate_boolean('upsert', upsert)

        filter, update = command_coder(filter, update,
                                       codec_op=self._database.codec_options)
        if array_filters:
            array_filters = command_coder(
                *array_filters, codec_op=self._database.codec_options)

        if bypass_document_validation:
            pass

        raw_result = {"n": 0, "nModified": 0}
        updator = Updator(update, array_filters)
        try:
            fw = next(self._internal_scan_query(filter))
        except StopIteration:
            if upsert:
                self._internal_upsert(filter, updator, raw_result)
        else:
            raw_result["n"] = 1
            if updator(fw):
                self._storage.update_one(self, fw.doc)
                raw_result["nModified"] = 1

        return UpdateResult(raw_result)

    def update_many(self,
                    filter,
                    update,
                    upsert=False,
                    bypass_document_validation=False,
                    array_filters=None, *args, **kwargs):
        """
        """
        validate_is_mapping("filter", filter)
        validate_ok_for_update(update)
        validate_list_or_none('array_filters', array_filters)
        validate_boolean('upsert', upsert)

        filter, update = command_coder(filter, update,
                                       codec_op=self._database.codec_options)
        if array_filters:
            array_filters = command_coder(
                *array_filters, codec_op=self._database.codec_options)

        if bypass_document_validation:
            pass

        raw_result = {"n": 0, "nModified": 0}
        updator = Updator(update, array_filters)
        scanner = self._internal_scan_query(filter)
        try:
            next(scanner)
        except StopIteration:
            if upsert:
                self._internal_upsert(filter, updator, raw_result)
        else:
            def update_docs():
                n, m = 0, 0
                for fieldwalker in scanner:
                    n += 1
                    if updator(fieldwalker):
                        m += 1
                        yield fieldwalker.doc
                raw_result["n"] = n
                raw_result["nModified"] = m

            self._storage.update_many(self, update_docs())

        return UpdateResult(raw_result)

    def delete_one(self, filter):
        raw_result = {"n": 0}

        queryfilter = QueryFilter(filter)
        storage = self._storage
        documents = storage.query(MontyCursor(self), 0)

        for doc in documents:
            if queryfilter(doc):
                storage.delete_one(self, doc["_id"])
                raw_result["n"] = 1
                break

        return DeleteResult(raw_result)

    def delete_many(self, filter):
        raw_result = {"n": 0}

        queryfilter = QueryFilter(filter)
        storage = self._storage
        documents = storage.query(MontyCursor(self), 0)

        doc_ids = set()
        for doc in documents:
            if queryfilter(doc):
                doc_ids.add(doc["_id"])
                raw_result["n"] += 1

        storage.delete_many(self, doc_ids)

        return DeleteResult(raw_result)

    def aggregate(self, pipeline, session=None, **kwargs):
        # return CommandCursor
        raise NotImplementedError("Not implemented.")

    def aggregate_raw_batches(self, pipeline, **kwargs):
        # return RawBatchCursor
        raise NotImplementedError("Not implemented.")

    def watch(self,
              pipeline=None,
              full_document='default',
              resume_after=None,
              max_await_time_ms=None,
              batch_size=None,
              collation=None,
              session=None):
        # return a changeStreamCursor
        raise NotImplementedError("Not implemented.")

    def find(self, *args, **kwargs):
        # return a cursor
        return MontyCursor(self, *args, **kwargs)

    def find_raw_batches(self, *args, **kwargs):
        # return RawBatchCursor
        raise NotImplementedError("Not implemented.")

    def find_one(self, filter=None, *args, **kwargs):
        """
        """
        if (filter is not None and not
                isinstance(filter, collections.Mapping)):
            filter = {"_id": filter}

        cursor = self.find(filter, *args, **kwargs)
        for result in cursor.limit(-1):
            return result
        return None

    def find_one_and_delete(self, filter, projection=None, sort=None):
        raise NotImplementedError("Not implemented.")

    def find_one_and_replace(self,
                             filter,
                             replacement,
                             projection=None,
                             sort=None, *args, **kwargs):
        raise NotImplementedError("Not implemented.")

    def find_one_and_update(self,
                            filter,
                            update,
                            projection=None,
                            sort=None, *args, **kwargs):
        raise NotImplementedError("Not implemented.")

    def count(self, filter=None):
        warnings.warn("count is deprecated. Use Collection.count_documents "
                      "instead.", DeprecationWarning, stacklevel=2)

        documents = self._storage.query(MontyCursor(self), 0)
        if filter:
            return self.__count_with_filter(documents, filter)
        else:
            return len(list(documents))

    def count_documents(self, filter):
        documents = self._storage.query(MontyCursor(self), 0)
        return self.__count_with_filter(documents, filter)

    def __count_with_filter(self, documents, filter):
        count = 0
        queryfilter = QueryFilter(filter)
        for doc in documents:
            if queryfilter(doc):
                count += 1
        return count

    def distinct(self, key, filter=None, **kwargs):
        """
        """
        if not isinstance(key, string_type):
            raise TypeError("key must be an "
                            "instance of %s" % (string_type.__name__,))

        result = list()

        def get_value(doc):
            fieldwalker = FieldWalker(doc)
            fieldvalues = fieldwalker.go(key).get().value
            res = list()
            for v in fieldvalues.iter_flat():
                weighted = Weighted(v)
                if weighted not in result:
                    res.append(weighted)
            return sorted(res)

        documents = self._storage.query(MontyCursor(self), 0)

        if filter:
            queryfilter = QueryFilter(filter)
            for doc in documents:
                if queryfilter(doc):
                    result += get_value(doc)
        else:
            for doc in documents:
                result += get_value(doc)

        return [weighted.value for weighted in result]

    def create_index(self, keys):
        raise NotImplementedError("Not implemented.")

    def create_indexes(self, indexes):
        raise NotImplementedError("Not implemented.")

    def drop_index(self, index_or_name):
        raise NotImplementedError("Not implemented.")

    def drop_indexes(self):
        raise NotImplementedError("Not implemented.")

    def reindex(self):
        raise NotImplementedError("Not implemented.")

    def list_indexes(self):
        # return a commandCursor
        raise NotImplementedError("Not implemented.")

    def index_information(self):
        raise NotImplementedError("Not implemented.")

    def drop(self):
        self._database.drop_collection(self._name)

    def save(self, to_save, *args, **kwargs):
        # DEPRECATED
        if "_id" in to_save:
            self.replace_one({"_id": to_save["_id"]},
                             to_save,
                             upsert=True,
                             *args, **kwargs)
        else:
            self.insert_one(to_save, *args, **kwargs)

    def rename(self, new_name, session=None, **kwargs):
        raise NotImplementedError("Not implemented.")

    def options(self, session=None):
        raise NotImplementedError("Not implemented.")

    def map_reduce(self,
                   map,
                   reduce,
                   out,
                   full_response=False,
                   session=None, **kwargs):
        raise NotImplementedError("Not implemented.")

    def inline_map_reduce(self,
                          map,
                          reduce,
                          full_response=False,
                          session=None, **kwargs):
        raise NotImplementedError("Not implemented.")

    def parallel_scan(self, num_cursors, session=None, **kwargs):
        raise NotImplementedError("Not implemented.")




from collections import OrderedDict
from datetime import datetime

from bson.py3compat import string_type
from bson.decimal128 import Decimal128
from bson.timestamp import Timestamp

from ..errors import WriteError
from .core import (
    _cmp_decimal,
    FieldWalker,
    Weighted,
    FieldWriteError,
)
from .queries import QueryFilter, ordering
from .helpers import is_numeric_type, is_duckument_type, is_integer_type


def _update(fieldwalker,
            field,
            value,
            evaluator,
            array_filters):

    fieldwalker.go(field)
    try:
        fieldwalker.set(value, evaluator, array_filters)
    # Take error message and put error code
    except FieldWriteError as err:
        raise WriteError(str(err), code=err.code)


def _drop(fieldwalker, field, array_filters):

    fieldwalker.go(field)
    try:
        fieldwalker.drop(array_filters)
    # Take error message and put error code
    except FieldWriteError as err:
        raise WriteError(str(err), code=err.code)


class Updator(object):

    def __init__(self, spec, array_filters=None):

        self.update_ops = {

            # field update ops
            "$inc": parse_inc,
            "$min": parse_min,
            "$max": parse_max,
            "$mul": parse_mul,
            "$rename": parse_rename,
            "$set": parse_set,
            "$setOnInsert": self.parse_set_on_insert,
            "$unset": parse_unset,
            "$currentDate": parse_currentDate,

            # array update ops
            # $                 implemented in FieldWalker
            # $[]               implemented in FieldWalker
            # $[<identifier>]   implemented in FieldWalker
            "$addToSet": parse_add_to_set,
            "$pop": parse_pop,
            "$pull": parse_pull,
            "$push": parse_push,
            "$pullAll": parse_pull_all,

            # $each             implemented in Eacher
            # $position         implemented in Eacher
            # $slice            implemented in Eacher
            # $sort             implemented in Eacher
        }

        self.fields_to_update = []
        self.array_filters = self.array_filter_parser(array_filters or [])
        # sort by key (operator)
        self.operations = OrderedDict(sorted(self.parser(spec).items()))
        self.__insert = None
        self.__fieldwalker = None

    def __repr__(self):
        pass

    def __call__(self, fieldwalker, do_insert=False):
        """Update document and return a bool value indicate changed or not"""
        self.__fieldwalker = fieldwalker
        self.__insert = do_insert

        with fieldwalker:
            for operator in self.operations.values():
                operator(fieldwalker)

            return fieldwalker.commit()

    @property
    def fieldwalker(self):
        return self.__fieldwalker

    def array_filter_parser(self, array_filters):
        filters = {}
        for i, filter_ in enumerate(array_filters):
            top = ""
            conds = {}

            for identifier, cond in filter_.items():
                id_s = identifier.split(".", 1)

                if not top and id_s[0] in filters:
                    msg = ("Found multiple array filters with the same "
                           "top-level field name {}".format(id_s[0]))
                    raise WriteError(msg, code=9)

                if top and id_s[0] != top:
                    msg = ("Error parsing array filter: Expected a single "
                           "top-level field name, found {0!r} and {1!r}"
                           "".format(top, id_s[0]))
                    raise WriteError(msg, code=9)

                top = id_s[0]
                conds.update({identifier: cond})

            filters[top] = QueryFilter(conds)

        return filters

    def parser(self, spec):
        if not next(iter(spec)).startswith("$"):
            raise ValueError("update only works with $ operators")

        update_stack = {}
        idnt_tops = list(self.array_filters.keys())

        for op, cmd_doc in spec.items():
            if op not in self.update_ops:
                raise WriteError("Unknown modifier: {}".format(op))

            if not is_duckument_type(cmd_doc):
                msg = ("Modifiers operate on fields but we found type {0} "
                       "instead. For example: {{$mod: {{<field>: ...}}}} "
                       "not {1}".format(type(cmd_doc).__name__, spec))
                raise WriteError(msg, code=9)

            for field, value in cmd_doc.items():
                if field == "_id":
                    msg = ("Performing an update on the path '_id' would "
                           "modify the immutable field '_id'")
                    raise WriteError(msg, code=66)

                for top in list(idnt_tops):
                    if "$[{}]".format(top) in field:
                        idnt_tops.remove(top)

                update_stack[field] = self.update_ops[op](
                    field, value, self.array_filters)

                self.check_conflict(field)
                if op == "$rename":
                    self.check_conflict(value)

        if idnt_tops:
            msg = ("The array filter for identifier {0!r} was not "
                   "used in the update {1}".format(idnt_tops[0], spec))
            raise WriteError(msg, code=9)

        return update_stack

    def check_conflict(self, field):
        for staged in self.fields_to_update:
            if field.startswith(staged) or staged.startswith(field):
                msg = ("Updating the path {0!r} would create a "
                       "conflict at {1!r}".format(field, staged[:len(field)]))
                raise WriteError(msg, code=40)

        self.fields_to_update.append(field)

    def parse_set_on_insert(self, field, value, array_filters):
        def _set_on_insert(fieldwalker):
            if self.__insert:
                parse_set(field, value, array_filters)(fieldwalker)

        return _set_on_insert


def parse_inc(field, value, array_filters):
    if not is_numeric_type(value):
        val_repr_ = "{!r}" if isinstance(value, string_type) else "{}"
        val_repr_ = val_repr_.format(value)
        msg = ("Cannot increment with non-numeric argument: "
               "{{{0}: {1}}}".format(field, val_repr_))
        raise WriteError(msg, code=14)

    def _inc(fieldwalker):

        def evaluator(node, inc_val):
            old_val = node.value
            if node.exists and not is_numeric_type(old_val):
                _id = fieldwalker.doc["_id"]
                value_type = type(old_val).__name__
                msg = ("Cannot apply $inc to a value of non-numeric type. "
                       "{{_id: {0}}} has the field {1!r} of non-numeric type "
                       "{2}".format(_id, str(node), value_type))
                raise WriteError(msg, code=14)

            is_decimal128 = False
            if isinstance(old_val, Decimal128):
                is_decimal128 = True
                old_val = old_val.to_decimal()
            if isinstance(inc_val, Decimal128):
                is_decimal128 = True
                inc_val = inc_val.to_decimal()

            if is_decimal128:
                return Decimal128((old_val or 0) + inc_val)
            else:
                return (old_val or 0) + inc_val

        _update(fieldwalker, field, value, evaluator, array_filters)

    return _inc


def parse_min(field, value, array_filters):

    def _min(fieldwalker):

        def evaluator(node, min_val):
            old_val = node.value
            if node.exists:
                old_val = Weighted(old_val)
                min_val = Weighted(min_val)
                return min_val.value if min_val < old_val else old_val.value
            else:
                return min_val

        _update(fieldwalker, field, value, evaluator, array_filters)

    return _min


def parse_max(field, value, array_filters):

    def _max(fieldwalker):

        def evaluator(node, max_val):
            old_val = node.value
            if node.exists:
                old_val = Weighted(old_val)
                max_val = Weighted(max_val)
                return max_val.value if max_val > old_val else old_val.value
            else:
                return max_val

        _update(fieldwalker, field, value, evaluator, array_filters)

    return _max


def parse_mul(field, value, array_filters):
    if not is_numeric_type(value):
        val_repr_ = "{!r}" if isinstance(value, string_type) else "{}"
        val_repr_ = val_repr_.format(value)
        msg = ("Cannot multiply with non-numeric argument: "
               "{{{0}: {1}}}".format(field, val_repr_))
        raise WriteError(msg, code=14)

    def _mul(fieldwalker):

        def evaluator(node, mul_val):
            old_val = node.value
            if node.exists and not is_numeric_type(old_val):
                _id = fieldwalker.doc["_id"]
                value_type = type(old_val).__name__
                msg = ("Cannot apply $mul to a value of non-numeric type. "
                       "{{_id: {0}}} has the field {1!r} of non-numeric type "
                       "{2}".format(_id, str(node), value_type))
                raise WriteError(msg, code=14)

            is_decimal128 = False
            if isinstance(old_val, Decimal128):
                is_decimal128 = True
                old_val = old_val.to_decimal()
            if isinstance(mul_val, Decimal128):
                is_decimal128 = True
                mul_val = mul_val.to_decimal()

            if is_decimal128:
                return Decimal128((old_val or 0) * mul_val)
            else:
                return (old_val or 0.0) * mul_val

        _update(fieldwalker, field, value, evaluator, array_filters)

    return _mul


def _get_array_member(fieldvalues):
    for node in fieldvalues.nodes:
        if node.in_array:
            return node


def parse_rename(field, new_field, array_filters):
    if not isinstance(new_field, string_type):
        msg = ("The 'to' field for $rename must be a string: {0}: {1}"
               "".format(field, new_field))
        raise WriteError(msg, code=2)

    if field == new_field:
        msg = ("The source and target field for $rename must differ: "
               "{0}: {1!r}".format(field, new_field))
        raise WriteError(msg, code=2)

    if field.startswith(new_field) or new_field.startswith(field):
        msg = ("The source and target field for $rename must not be on the "
               "same path: {0}: {1!r}".format(field, new_field))
        raise WriteError(msg, code=2)

    def _rename(fieldwalker):

        probe = FieldWalker(fieldwalker.doc)

        probe.go(field).get()
        fieldvalues = probe.value

        if not fieldvalues.is_exists():
            return

        value = next(fieldvalues.iter_plain())

        array_member = _get_array_member(fieldvalues)
        if array_member is not None:
            _id = probe.doc["_id"]
            array_field = str(array_member.parent)
            msg = ("The source field cannot be an array element, "
                   "{0!r} in doc with _id: {1} has an array field "
                   "called {2!r}".format(field, _id, array_field))
            raise WriteError(msg, code=2)

        probe.go(new_field).get()
        fieldvalues = probe.value

        array_member = _get_array_member(fieldvalues)
        if array_member is not None:
            _id = probe.doc["_id"]
            array_field = str(array_member.parent)
            msg = ("The destination field cannot be an array element, "
                   "{0!r} in doc with _id: {1} has an array field "
                   "called {2!r}".format(new_field, _id, array_field))
            raise WriteError(msg, code=2)

        _drop(fieldwalker, field, array_filters)
        _update(fieldwalker, new_field, value, None, array_filters)

    return _rename


def parse_set(field, value, array_filters):

    def _set(fieldwalker):

        _update(fieldwalker, field, value, None, array_filters)

    return _set


def parse_unset(field, _, array_filters):

    def _unset(fieldwalker):

        _drop(fieldwalker, field, array_filters)

    return _unset


def parse_currentDate(field, value, array_filters):
    date_type = {
        "date": datetime.utcnow(),
        "timestamp": Timestamp(datetime.utcnow(), 1),
    }

    if not isinstance(value, bool):
        if not is_duckument_type(value):
            msg = ("{} is not valid type for $currentDate. Please use a "
                   "boolean ('true') or a $type expression ({{$type: "
                   "'timestamp/date'}}).".format(type(value).__name__))
            raise WriteError(msg, code=2)

        for k, v in value.items():
            if k != "$type":
                msg = "Unrecognized $currentDate option: {}".format(k)
                raise WriteError(msg, code=2)
            if v not in date_type:
                msg = ("The '$type' string field is required to be 'date' "
                       "or 'timestamp': {$currentDate: {field : {$type: "
                       "'date'}}}")
                raise WriteError(msg, code=2)

            value = date_type[v]
    else:
        value = date_type["date"]

    def _currentDate(fieldwalker):
        parse_set(field, value, array_filters)(fieldwalker)

    return _currentDate


def parse_add_to_set(field, value_or_each, array_filters):
    if (is_duckument_type(value_or_each) and
            next(iter(value_or_each)) == "$each"):
        value = EachAdder(value_or_each)
        run_each = True
    else:
        value = value_or_each
        run_each = False

    def _add_to_set(fieldwalker):

        def evaluator(node, new_elem):
            old_val = node.value
            if node.exists and not isinstance(old_val, list):
                value_type = type(old_val).__name__
                msg = ("Cannot apply $addToSet to non-array field. Field "
                       "named {0!r} has non-array type {1}"
                       "".format(str(node), value_type))
                raise WriteError(msg, code=2)

            if run_each:
                eacher = new_elem
                new_array = eacher(old_val)
            else:
                new_array = (old_val or [])[:]
                if new_elem not in new_array:
                    new_array.append(new_elem)

            return new_array

        _update(fieldwalker, field, value, evaluator, array_filters)

    return _add_to_set


def parse_pop(field, value, array_filters):
    if not is_numeric_type(value):
        msg = ("Expected a number in: {0}: {1!r}".format(field, value))
        raise WriteError(msg, code=9)
    else:
        try:
            value = float(value)
            msg_raw = "Expected an integer: {0}: {1!r}"
        except TypeError:
            msg_raw = "Cannot represent as a 64-bit integer: {0}: {1!r}"
            value = float(value.to_decimal())

        if value not in (1.0, -1.0):
            raise WriteError(msg_raw.format(field, value), code=9)

    def _pop(fieldwalker):

        def evaluator(node, pop_ind):
            old_val = node.value
            if node.exists and not isinstance(old_val, list):
                value_type = type(old_val).__name__
                msg = ("Path {0!r} contains an element of non-array type "
                       "{1!r}".format(str(node), value_type))
                raise WriteError(msg, code=14)

            if not node.exists:
                # do nothing
                return old_val

            if pop_ind == 1:
                return old_val[:-1]
            else:
                return old_val[1:]

        _update(fieldwalker, field, value, evaluator, array_filters)

    return _pop


def parse_pull(field, value_or_conditions, array_filters):
    if is_duckument_type(value_or_conditions):
        query_spec = {}
        for k, v in value_or_conditions.items():
            if not k[:1] == "$":
                query_spec[".".join((field, k))] = v
            else:
                query_spec[field] = {k: v}
        queryfilter = QueryFilter(query_spec)
    else:
        queryfilter = QueryFilter({field: value_or_conditions})

    def _pull(fieldwalker):

        def evaluator(node, _):
            old_val = node.value
            if node.exists and not isinstance(old_val, list):
                msg = "Cannot apply $pull to a non-array value"
                raise WriteError(msg, code=2)

            if not node.exists:
                # do nothing
                return old_val

            new_array = []
            for elem in old_val:
                result = queryfilter({field: elem})

                if not result:
                    new_array.append(elem)
            return new_array

        _update(fieldwalker, field, None, evaluator, array_filters)

    return _pull


def parse_push(field, value_or_each, array_filters):
    if is_duckument_type(value_or_each) and "$each" in value_or_each:
        value = EachPusher(value_or_each)
        run_each = True
    else:
        value = value_or_each
        run_each = False

    def _push(fieldwalker):

        def evaluator(node, new_elem):
            old_val = node.value
            if node.exists and not isinstance(old_val, list):
                value_type = type(old_val).__name__
                _id = fieldwalker.doc["_id"]
                msg = ("The field {0!r} must be an array but is of type "
                       "{1} in document {{_id: {2}}}"
                       "".format(str(node), value_type, _id))
                raise WriteError(msg, code=2)

            if run_each:
                eacher = new_elem
                new_array = eacher(old_val)
            else:
                new_array = (old_val or [])[:]
                new_array.append(new_elem)

            return new_array

        _update(fieldwalker, field, value, evaluator, array_filters)

    return _push


def parse_pull_all(field, value, array_filters):
    if not isinstance(value, list):
        value_type = type(value).__name__
        msg = ("$pullAll requires an array argument but was given a {}"
               "".format(value_type))
        raise WriteError(msg, code=2)

    def _pull_all(fieldwalker):

        def evaluator(node, pull_list):
            old_val = node.value
            if node.exists and not isinstance(old_val, list):
                msg = "Cannot apply $pull to a non-array value"
                raise WriteError(msg, code=2)

            if not node.exists:
                # do nothing
                return old_val

            def convert(lst):
                for val in lst:
                    if isinstance(val, Decimal128):
                        yield _cmp_decimal(val)
                    else:
                        yield val

            pull_list = list(convert(pull_list))
            old_val = list(convert(old_val))

            new_array = [elem for elem in old_val if elem not in pull_list]
            return new_array

        _update(fieldwalker, field, value, evaluator, array_filters)

    return _pull_all


class EachAdder(object):

    def __init__(self, spec):
        spec = spec.copy()

        self.mods = {
            "$each": None,
        }

        for mod, value in spec.items():
            try:
                type_check = self.validators[mod]
            except KeyError:
                raise WriteError("Found unexpected fields after $each in "
                                 "$addToSet: %s" % spec.to_dict(),  # SON type
                                 code=2)

            self.mods[mod] = type_check(self, value)

    def __call__(self, array):
        new_array = (array or [])[:]
        new_elems = self.mods["$each"][:]

        new_array[0:] += [e for e in new_elems if e not in new_array]
        return new_array

    def _validate_each(self, each):
        try:
            each[:]
        except TypeError:
            type_name = type(each).__name__
            raise WriteError("The argument to $each in $addToSet must be an "
                             "array but it was of type %s" % type_name,
                             code=14)
        return each

    validators = {
        "$each": _validate_each,
    }


class EachPusher(object):

    def __init__(self, spec):
        spec = spec.copy()

        self.mods = {
            "$each": None,
            "$position": None,
            "$slice": None,
            "$sort": None,
        }

        for mod, value in spec.items():
            try:
                type_check = self.validators[mod]
            except KeyError:
                raise WriteError("Unrecognized clause in $push: %s" % mod,
                                 code=2)

            self.mods[mod] = type_check(self, value)

    def __call__(self, array):
        new_array = (array or [])[:]
        new_elems = self.mods["$each"][:]

        position = self.mods["$position"]
        slice = self.mods["$slice"]
        sort = self.mods["$sort"]

        if position is None:
            new_array += new_elems
        else:
            new_array[:position] += new_elems

        if slice is not None:
            if slice >= 0:
                new_array = new_array[:slice]
            else:
                new_array = new_array[slice:]

        if sort is not None:
            if is_duckument_type(sort):
                fieldwalkers = list()
                unsortable = list()
                for elem in new_array:
                    if is_duckument_type(elem):
                        fieldwalkers.append(FieldWalker(elem))
                    else:
                        unsortable.append(elem)

                ordered = ordering(fieldwalkers, sort)
                new_array = [f.doc for f in ordered]

                if unsortable:
                    is_reverse = bool(1 - next(iter(sort.values())))
                    if is_reverse:
                        new_array += unsortable
                    else:
                        new_array[:0] += unsortable

            else:
                is_reverse = bool(1 - sort)
                ordered = sorted((Weighted(e) for e in new_array),
                                 reverse=is_reverse)
                new_array = [w.value for w in ordered]

        return new_array

    def _validate_each(self, each):
        try:
            each[:]
        except TypeError:
            type_name = type(each).__name__
            raise WriteError("The argument to $each in $push must be an "
                             "array but it was of type: %s" % type_name,
                             code=2)
        return each

    def _validate_position(self, position):
        if not is_integer_type(position):
            type_name = type(position).__name__
            raise WriteError("The value for $position must be an integer "
                             "value, not of type: %s" % type_name,
                             code=2)
        return position

    def _validate_slice(self, slice):
        if not is_integer_type(slice):
            type_name = type(slice).__name__
            raise WriteError("The value for $slice must be an integer "
                             "value but was given type: %s" % type_name,
                             code=2)
        return slice

    def _validate_sort(self, sort, int_only=False):
        if is_integer_type(sort) or int_only:
            if sort not in (1, -1):
                raise WriteError("The $sort element value must be either "
                                 "1 or -1",
                                 code=2)
            return sort

        if is_duckument_type(sort):
            for key, value in sort.items():
                self._validate_sort(value, int_only=True)
            return sort

        raise WriteError("The $sort is invalid: use 1/-1 to sort the whole "
                         "element, or {field:1/-1} to sort embedded fields",
                         code=2)

    validators = {
        "$each": _validate_each,
        "$position": _validate_position,
        "$slice": _validate_slice,
        "$sort": _validate_sort,
    }




import os
import shutil
import sqlite3
import contextlib
from collections import OrderedDict

from bson import BSON
from bson.py3compat import _unicode

from ..base import WriteConcern
from .abcs import (
    AbstractStorage,
    AbstractDatabase,
    AbstractCollection,
    AbstractCursor,
)
from .errors import StorageDuplicateKeyError


sqlite_324 = sqlite3.sqlite_version_info >= (3, 24, 0)


"""
(NOTE) SQLite3 pragmas DEFAULT value:

* Applies to the database
    - journal_mode=delete

* Applies to the connection
    - busy_timeout=5000
    - foreign_keys=OFF
    - automatic_index=ON
"""


SQLITE_DB_EXT = ".collection"
SQLITE_RECORD_TABLE = "documents"


"""SQL"""

CREATE_TABLE = """
    CREATE TABLE [{}](
        k text NOT NULL,
        v blob NOT NULL,
        PRIMARY KEY(k)
    );
"""

INSERT_RECORD = """
    INSERT INTO [{}](k, v) VALUES (?, ?);
"""

UPSERT_RECORD = """
    INSERT INTO [{}] (v, k) VALUES(?, ?)
        ON CONFLICT(k)
        DO UPDATE SET v=excluded.v;
"""  # need sqlite_version >= 3.24.0

UPDATE_RECORD = """
    UPDATE [{}] SET v = (?) WHERE k = (?);
"""

INSORE_RECORD = """
    INSERT OR IGNORE INTO [{}](v, k) VALUES (?, ?);
"""  # for sqlite_version < 3.24, UPDATE + INSERT_IGNORE = UPSERT SCRIPT

DELETE_RECORD = """
    DELETE FROM [{}] WHERE k = (?);
"""

SELECT_ALL_RECORD = """
    SELECT v FROM [{}];
"""

SELECT_LIMIT_RECORD = """
    SELECT v FROM [{0}] LIMIT {1};
"""

SELECT_ALL_KEYS = """
    SELECT k FROM [{}];
"""


class SQLiteKVEngine(object):

    def __init__(self, db_pragmas):
        self.__db_pragmas = db_pragmas
        self.__conn = None

    @property
    def db_pragmas(self):
        return self._assemble_pragmas(self.__db_pragmas)

    def _connect(self, db_file, wconcern=None):
        self.__conn = sqlite3.connect(db_file)
        self.__conn.text_factory = str

        wcon_pragmas = ""
        if wconcern:
            wcon_doc = wconcern.document
            # wtimeout (milliseconds) -> busy_timeout (milliseconds)
            timeout = wcon_doc.get("wtimeout")
            if timeout:
                del wcon_doc["wtimeout"]
                wcon_doc["busy_timeout"] = timeout

            wcon_pragmas = self._assemble_pragmas(wcon_doc)

        # update connection pragmas and write_concern pragmas
        self.__conn.executescript(self.db_pragmas + ";" + wcon_pragmas)

        return contextlib.closing(self.__conn)

    def _assemble_pragmas(self, pragma_dict):
        return ";".join(["PRAGMA {0}={1}".format(k, v)
                         for k, v in pragma_dict.items()])

    def create_table(self, db_file):
        with self._connect(db_file) as conn:
            with conn:
                conn.execute(CREATE_TABLE.format(SQLITE_RECORD_TABLE))

    def write_one(self, db_file, params, wconcern=None):
        with self._connect(db_file, wconcern) as conn:
            with conn:
                sql = INSERT_RECORD.format(SQLITE_RECORD_TABLE)
                conn.execute(sql, params)

    def write_many(self, db_file, seq_params, wconcern=None):
        with self._connect(db_file, wconcern) as conn:
            with conn:
                sql = INSERT_RECORD.format(SQLITE_RECORD_TABLE)
                conn.executemany(sql, seq_params)

    def update_one(self, db_file, params, wconcern=None):
        with self._connect(db_file, wconcern) as conn:
            with conn:
                sql = UPDATE_RECORD.format(SQLITE_RECORD_TABLE)
                conn.execute(sql, params)

    def update_many(self, db_file, seq_params, wconcern=None):
        with self._connect(db_file, wconcern) as conn:
            with conn:
                sql = UPDATE_RECORD.format(SQLITE_RECORD_TABLE)
                conn.executemany(sql, seq_params)

    def delete_one(self, db_file, params, wconcern=None):
        with self._connect(db_file, wconcern) as conn:
            with conn:
                sql = DELETE_RECORD.format(SQLITE_RECORD_TABLE)
                conn.execute(sql, params)

    def delete_many(self, db_file, seq_params, wconcern=None):
        with self._connect(db_file, wconcern) as conn:
            with conn:
                sql = DELETE_RECORD.format(SQLITE_RECORD_TABLE)
                conn.executemany(sql, seq_params)

    def read_all(self, db_file, limit):
        if not os.path.isfile(db_file):
            return []
        with self._connect(db_file) as conn:
            with conn:
                if limit:
                    sql = SELECT_LIMIT_RECORD.format(
                        SQLITE_RECORD_TABLE, limit)
                else:
                    sql = SELECT_ALL_RECORD.format(SQLITE_RECORD_TABLE)

                return conn.execute(sql).fetchall()

    def read_all_keys(self, db_file):
        if not os.path.isfile(db_file):
            return []
        with self._connect(db_file) as conn:
            with conn:
                sql = SELECT_ALL_KEYS.format(SQLITE_RECORD_TABLE)
                return conn.execute(sql).fetchall()


class SQLiteWriteConcern(WriteConcern):
    """

    Args:
        busy_timeout (int): Default 5000

        synchronous (int, str): Default "NORMAL"
            - type: integer
              enum: [0, 1, 2, 3]
            - type: string
              enum: ["OFF", NORMAL, FULL, EXTRA, "0", "1", "2", "3"]

        automatic_index (bool, str): Default False
            - type: boolean
            - type: string
              enum: ["ON", "OFF"]

    """

    def __init__(self,
                 busy_timeout=5000,
                 synchronous="NORMAL",
                 automatic_index=False):

        super(SQLiteWriteConcern, self).__init__(busy_timeout)

        if synchronous is not None:
            self._document["synchronous"] = synchronous
        if automatic_index is not None:
            self._document["automatic_index"] = automatic_index

    def __repr__(self):
        return ("SQLiteWriteConcern({})".format(
            ", ".join("%s=%s" % kvt for kvt in self.document.items()),))


class SQLiteStorage(AbstractStorage):
    """
    """

    def __init__(self, repository, storage_config):
        super(SQLiteStorage, self).__init__(repository, storage_config)
        self._conn = SQLiteKVEngine(self._config)

    def _db_path(self, db_name):
        """
        Get Monty database dir path.
        """
        return os.path.join(self._repository, db_name)

    @classmethod
    def nice_name(cls):
        return "sqlite"

    @classmethod
    def config(cls, journal_mode="WAL", **kwargs):
        """

        Args:
            journal_mode (str): Default "WAL"
                type: string
                enum: [DELETE, TRUNCATE, PERSIST, MEMORY, WAL, "OFF"]

        """
        return {
            "journal_mode": journal_mode,
        }

    def wconcern_parser(self,
                        wtimeout=None,
                        busy_timeout=None,
                        synchronous=None,
                        automatic_index=None,
                        **kwargs):
        return SQLiteWriteConcern(wtimeout or busy_timeout,
                                  synchronous,
                                  automatic_index)

    def database_create(self, db_name):
        if not os.path.isdir(self._db_path(db_name)):
            os.makedirs(self._db_path(db_name))

    def database_drop(self, db_name):
        db_path = self._db_path(db_name)
        if os.path.isdir(db_path):
            shutil.rmtree(db_path)

    def database_list(self):
        return [
            name for name in os.listdir(_unicode(self._repository))
            if os.path.isdir(self._db_path(name))
        ]


class SQLiteDatabase(AbstractDatabase):

    def __init__(self, storage, subject):
        super(SQLiteDatabase, self).__init__(storage, subject)
        self._db_path = storage._db_path(self._name)

    def _col_path(self, col_name):
        """
        Get SQLite database file path, which is Monty collection.
        """
        return os.path.join(self._db_path, col_name) + SQLITE_DB_EXT

    @property
    def _conn(self):
        return self._storage._conn

    def database_exists(self):
        return os.path.isdir(self._db_path)

    def collection_exists(self, col_name):
        return os.path.isfile(self._col_path(col_name))

    def collection_create(self, col_name):
        if not self.database_exists():
            self._storage.database_create(self._name)
        self._conn.create_table(self._col_path(col_name))

    def collection_drop(self, col_name):
        if self.collection_exists(col_name):
            os.remove(self._col_path(col_name))

    def collection_list(self):
        if not self.database_exists():
            return []
        return [os.path.splitext(name)[0]
                for name in os.listdir(_unicode(self._db_path))]


SQLiteStorage.contractor_cls = SQLiteDatabase


class SQLiteCollection(AbstractCollection):

    def __init__(self, database, subject):
        super(SQLiteCollection, self).__init__(database, subject)

        self._col_path = self._database._col_path(self._name)

    def _ensure_table(func):
        def make_table(self, *args, **kwargs):
            if not self._database.collection_exists(self._name):
                self._database.collection_create(self._name)
            return func(self, *args, **kwargs)
        return make_table

    def _encode_doc(self, doc, check_keys=False):
        # Preserve BSON types
        encoded = BSON.encode(doc,
                              # Check if keys start with '$' or contain '.'
                              check_keys=check_keys,
                              codec_options=self.coptions)
        return sqlite3.Binary(encoded)

    @property
    def _conn(self):
        return self._database._conn

    @_ensure_table
    def write_one(self, doc, check_keys=True):
        """
        """
        try:
            self._conn.write_one(
                self._col_path,
                (str(doc["_id"]), self._encode_doc(doc, check_keys),),
                self.wconcern
            )
        except sqlite3.IntegrityError:
            raise StorageDuplicateKeyError()

        return doc["_id"]

    @_ensure_table
    def write_many(self, docs, check_keys=True, ordered=True):
        """
        """
        _docs = OrderedDict()
        ids = list()
        keys = self._conn.read_all_keys(self._col_path)
        has_duplicated_key = False
        for doc in docs:
            id = doc["_id"]
            if str(id) in keys or str(id) in _docs:
                has_duplicated_key = True
                break

            _docs[str(id)] = self._encode_doc(doc, check_keys)
            ids.append(id)

        self._conn.write_many(
            self._col_path,
            _docs.items(),
            self.wconcern
        )

        if has_duplicated_key:
            raise StorageDuplicateKeyError()

        return ids

    def update_one(self, doc):
        """
        """
        self._conn.update_one(
            self._col_path,
            (self._encode_doc(doc), str(doc["_id"])),
            self.wconcern
        )

    def update_many(self, docs):
        """
        """
        self._conn.update_many(
            self._col_path,
            [(self._encode_doc(doc), str(doc["_id"])) for doc in docs],
            self.wconcern
        )

    def delete_one(self, id):
        self._conn.delete_one(
            self._col_path,
            (str(id),),
            self.wconcern
        )

    def delete_many(self, ids):
        self._conn.delete_many(
            self._col_path,
            [(str(id),) for id in ids],
            self.wconcern
        )


SQLiteDatabase.contractor_cls = SQLiteCollection


class SQLiteCursor(AbstractCursor):

    def __init__(self, collection, subject):
        super(SQLiteCursor, self).__init__(collection, subject)

    @property
    def _conn(self):
        return self._collection._conn

    @property
    def _col_path(self):
        return self._collection._col_path

    def query(self, max_scan):
        docs = self._conn.read_all(self._col_path, max_scan)
        return (self._decode_doc(doc[0]) for doc in docs)


SQLiteCollection.contractor_cls = SQLiteCursor




import os
import time
from collections import defaultdict, OrderedDict
from datetime import datetime

from bson import decode_all, BSON
from bson.codec_options import CodecOptions
from bson.py3compat import string_type
from bson.json_util import (
    loads as _loads,
    dumps as _dumps,
    RELAXED_JSON_OPTIONS as _default_json_opts,
)
from ..client import MontyClient
from ..errors import DuplicateKeyError


def _collection(database, collection):
    client = MontyClient()
    return client[database][collection]


def montyimport(database,
                collection,
                file,
                mode="insert",
                json_options=None):
    """Imports content from an Extended JSON file into a MontyCollection instance

    Example:
        >>> from montydb import open_repo, utils
        >>> with open_repo("foo/bar"):
        >>>     utils.montyimport("db", "col", "/data/dump.json")
        >>>

    Args:
        database (str): Database name
        collection (str): Collection name to import to
        file (str): Input file path
        mode (str): Specifies how the import process should handle existing
                    documents in the database that match documents in the
                    import file.
                    Options: ["insert", "upsert", "merge"]
                    Default: "insert"
        json_options (JSONOptions): A JSONOptions instance used to modify
                                    the decoding of MongoDB Extended JSON
                                    types. Default None.

    """
    collection = _collection(database, collection)
    opt = json_options or _default_json_opts

    with open(file, "r") as fp:
        lines = [line.strip() for line in fp.readlines()]
        serialized = "[{}]".format(", ".join(lines))

    documents = _loads(serialized, json_options=opt)

    if mode == "insert":
        for doc in documents:
            try:
                collection.insert_one(doc)
            except DuplicateKeyError:
                print("Duplicate id: %s" % doc["_id"])

    elif mode == "upsert":
        for doc in documents:
            collection.replace_one({"_id": doc["_id"]}, doc, upsert=True)

    elif mode == "merge":
        for doc in documents:
            update = {"$setOnInsert": {k: v} for k, v in doc.items()}
            collection.update_one({"_id": doc["_id"]}, update, upsert=True)


def montyexport(database,
                collection,
                out,
                fileds=None,
                query=None,
                json_options=None):
    """Produces a JSON export of data stored in a MontyCollection instance

    Example:
        >>> from montydb import open_repo, utils
        >>> with open_repo("foo/bar"):
        >>>     utils.montyexport("db", "col", "/data/dump.json")
        >>>

    Args:
        database (str): Database name
        collection (str): Collection name to export from
        out (str): Output file path
        fields (str, list): Specifies a field name string or a list fields
                            to include in the export.
        query (dict): Provides a query document to return matching documents
                      in the export.
        json_options (JSONOptions): A JSONOptions instance used to modify
                                    the decoding of MongoDB Extended JSON
                                    types. Default None.

    """
    collection = _collection(database, collection)
    opt = json_options or _default_json_opts
    fileds = fileds or []

    out = os.path.abspath(out)
    if not os.path.isdir(os.path.dirname(out)):
        os.makedirs(os.path.dirname(out))

    if isinstance(fileds, string_type):
        fileds = [fileds]

    projection = {field: True for field in fileds} or None

    with open(out, "w") as fp:
        for doc in collection.find(query, projection=projection):
            serialized = _dumps(doc, json_options=opt)
            fp.write(serialized + "\n")


def montyrestore(database, collection, dumpfile):
    """Loads a binary database dump into a MontyCollection instance

    Should be able to accept the dump created by `mongodump`.

    Example:
        >>> from montydb import open_repo, utils
        >>> with open_repo("foo/bar"):
        >>>     utils.montyrestore("db", "col", "/data/dump.bson")
        >>>

    Args:
        database (str): Database name
        collection (str): Collection name to load into
        dumpfile (str): File path to load from

    """
    collection = _collection(database, collection)

    with open(dumpfile, "rb") as fp:
        raw = fp.read()

    try:
        collection.insert_many(decode_all(raw))
    except DuplicateKeyError:
        pass


def montydump(database, collection, dumpfile):
    """Creates a binary export from a MontyCollection instance

    The export should be able to be accepted by `mongorestore`.

    Example:
        >>> from montydb import open_repo, utils
        >>> with open_repo("foo/bar"):
        >>>     utils.montydump("db", "col", "/data/dump.bson")
        >>>

    Args:
        database (str): Database name
        collection (str): Collection name to export from
        dumpfile (str): File path to export to

    """
    collection = _collection(database, collection)

    dumpfile = os.path.abspath(dumpfile)
    if not os.path.isdir(os.path.dirname(dumpfile)):
        os.makedirs(os.path.dirname(dumpfile))

    raw = b""
    for doc in collection.find():
        raw += BSON.encode(doc)

    with open(dumpfile, "wb") as fp:
        fp.write(raw)


class MongoQueryRecorder(object):
    """Record MongoDB query results in a period of time

    :Important: Requires to access databse profiler.

    This works via filtering the database profile data and reproduce the
    queries of `find` and `distinct` commands.

    Example:
        >>> from pymongo import MongoClient
        >>> from montydb.utils import MongoQueryRecorder
        >>> client = MongoClient()
        >>> recorder = MongoQueryRecorder(client["mydb"])
        >>> recorder.start()
        >>> # Make some queries or run the App...
        >>> recorder.stop()
        >>> recorder.extract()
        {<collection_1>: [<doc_1>, <doc_2>, ...], ...}

    Args:
        mongodb (pymongo.database.Database): An instance of mongo database
        namespace (str or regex, optional): A MongoDB namespace string/regex.
        user (str, optional): Name of authenticated user to record with.

    """

    def __init__(self, mongodb, namespace=None, user=None):
        self._mongodb = mongodb
        self._namespace = namespace or {"$regex": mongodb.name + "\..*"}
        self._user = user

        self._epoch = datetime(1970, 1, 1)
        self._rec_stime = None
        self._rec_etime = None

    def __repr__(self):
        return ("MongoQueryRecorder(mongodb=%s, namespace=%s, user=%s)"
                "" % (self._mongodb.name, self._namespace, self._user))

    def reset_profile(self, level=0):
        """Drop and reset database profile

        Args:
            level (int): Database profile level, default 0.

        """
        self._mongodb.command({"profile": 0})
        self._mongodb.system.profile.drop()
        if level:
            self._mongodb.command({"profile": level})

    def current_level(self):
        """Return current database's profile level"""
        return self._mongodb.command({"profile": -1})["was"]

    def start(self):
        """Start recording and set database profile level to 2"""
        self._mongodb.command({"profile": 2})
        self._rec_stime = datetime.utcnow()
        time.sleep(0.01)  # Wait for db

    def stop(self):
        """Stop recording and set database profile level to 0"""
        time.sleep(0.01)  # Wait for db
        self._rec_etime = datetime.utcnow()
        self._mongodb.command({"profile": 0})

    def extract(self):
        """Collect documents via previous queries

        Via filtering the `[databse].system.profile`, parsing previous
        commands to reproduce the query results.

        NOTE: Depend on the `namespace`, the result may across multiple
              collections.

        Returns:
            dict: A dict of {collection: list of documents}

        """
        filter = {
            "$or": [
                {
                    "op": "query",
                    "command.find": {"$exists": True},
                    "nreturned": {"$gte": 1}
                },
                {
                    "op": "command",
                    "command.distinct": {"$exists": True}
                },
            ],
            "ns": self._namespace,
            "ts": {"$gte": self._rec_stime, "$lte": self._rec_etime},
        }

        if self._user is not None:
            filter.update({"user": self._user})

        projection = {
            "op": 1,

            "command.find": 1,
            "command.filter": 1,
            "command.sort": 1,
            "command.limit": 1,

            "command.distinct": 1,
            "command.key": 1,
            "command.query": 1
        }

        profile = self._mongodb.system.profile
        code_opt = CodecOptions(document_class=OrderedDict)
        profile = profile.with_options(codec_options=code_opt)

        history = defaultdict(list)
        for log in profile.find(filter, projection=projection):
            op = log.pop("op")
            if log not in history[op]:
                history[op].append(log)

        documents = defaultdict(dict)

        # Query - find
        for cmd in (log["command"] for log in history["query"]):
            col = cmd["find"]
            filter = cmd["filter"]
            limit = cmd.get("limit", 0)
            sort = None

            if limit:
                sort = list()
                for k, v in cmd.get("sort", dict()):
                    sort.append((k, v))

            for doc in self._mongodb[col].find(filter, sort=sort, limit=limit):
                id = doc["_id"]
                if id not in documents[col]:
                    documents[col][id] = doc

        # Command - distinct
        for cmd in (log["command"] for log in history["command"]):
            col = cmd["distinct"]
            key = cmd["key"]
            query = cmd.get("query")

            for value in self._mongodb[col].distinct(key, query):
                doc = self._mongodb[col].find_one({key: value})
                id = doc["_id"]
                if id not in documents[col]:
                    documents[col][id] = doc

        # Done
        return {col: list(docs.values()) for col, docs in documents.items()}




from itertools import islice
from collections import OrderedDict

from .abcs import (
    AbstractStorage,
    AbstractDatabase,
    AbstractCollection,
    AbstractCursor,
)
from .errors import StorageDuplicateKeyError


_repo = OrderedDict()


class MemoryStorage(AbstractStorage):
    """
    """

    def __init__(self, repository, storage_config):
        super(MemoryStorage, self).__init__(repository, storage_config)
        self._repo = _repo

    @classmethod
    def config(cls, **storage_kwargs):
        return dict()

    def database_create(self, db_name):
        self._repo[db_name] = OrderedDict()

    def database_drop(self, db_name):
        if db_name in self._repo:
            del self._repo[db_name]

    def database_list(self):
        return list(self._repo.keys())


class MemoryDatabase(AbstractDatabase):
    """
    """

    @property
    def _db(self):
        return self._storage._repo[self._name]

    def db_exists(self):
        return self._name in self._storage._repo

    def collection_exists(self, col_name):
        if self.db_exists():
            return col_name in self._db
        return False

    def collection_create(self, col_name):
        if not self.db_exists():
            self._storage.database_create(self._name)
        self._db[col_name] = OrderedDict()

    def collection_drop(self, col_name):
        if self.collection_exists(col_name):
            del self._db[col_name]

    def collection_list(self):
        if not self.db_exists():
            return []
        return list(self._db.keys())


MemoryStorage.contractor_cls = MemoryDatabase


class MemoryCollection(AbstractCollection):
    """
    """

    @property
    def _col(self):
        if not self._col_exists():
            self._database.collection_create(self._name)
        return self._database._db[self._name]

    def _col_exists(self):
        return self._database.collection_exists(self._name)

    def _id_unique(self, id):
        if id in self._col:
            raise StorageDuplicateKeyError()

    def write_one(self, doc, check_keys=True):
        id = doc["_id"]
        self._id_unique(str(id))
        self._col[str(id)] = self._encode_doc(doc, check_keys)
        return id

    def write_many(self, docs, check_keys=True, ordered=True):
        ids = list()
        for doc in docs:
            id = doc["_id"]
            self._id_unique(str(id))
            self._col[str(id)] = self._encode_doc(doc, check_keys)
            ids.append(id)
        return ids

    def update_one(self, doc):
        self._col[str(doc["_id"])] = self._encode_doc(doc)

    def update_many(self, docs):
        for doc in docs:
            self._col[str(doc["_id"])] = self._encode_doc(doc)

    def delete_one(self, id):
        del self._col[str(id)]

    def delete_many(self, ids):
        for id in ids:
            del self._col[str(id)]


MemoryDatabase.contractor_cls = MemoryCollection


class MemoryCursor(AbstractCursor):
    """
    """

    @property
    def _col(self):
        if self._collection._col_exists():
            return self._collection._col
        return OrderedDict()

    def query(self, max_scan):
        docs = (self._decode_doc(doc) for doc in self._col.values())
        if not max_scan:
            return docs
        else:
            return islice(docs, max_scan)


MemoryCollection.contractor_cls = MemoryCursor



import platform
import sys

from bson.py3compat import string_type

from . import errors, version
from .base import BaseObject, ClientOptions
from .configure import provide_storage, provide_repository
from .database import MontyDatabase


class MontyClient(BaseObject):

    def __init__(self,
                 repository=None,
                 document_class=dict,
                 tz_aware=None,
                 **kwargs):
        """Client for a MontyDB instance

        The `repository` argument can be a montydb URI. A montydb URI is simply
        a repository path that prefixed with montydb scheme.
        For example:
            `montydb:///foo/bar/db_repo` -> Point to a absolute dir path
            `montydb://db_repo` -> Point to a relative dir path
            `montydb://` -> Point to current working dir or pinned dir
            `montydb://:memory:` -> Use memory storage

        Args:
            repository (str): A dir path for on-disk storage or `:memory:` for
                memory storage only, or a montydb URI.
            document_class (cls, optional): default class to use for documents
                returned from queries on this client. Default class is `dict`.
            tz_aware (bool, optional): if `True`, `datetime.datetime` instances
                returned as values in document by this client will be timezone
                awared (otherwise they will be naive).
            **kwargs: Other optional keyword arguments will pass into storage
                engine as write concern arguments.

        """
        repository = provide_repository(repository)
        storage_cls = provide_storage(repository)
        storage_instance = storage_cls.launch(repository)

        self._storage = storage_instance
        wconcern = self._storage.wconcern_parser(**kwargs)

        options = kwargs
        options["document_class"] = document_class
        options["tz_aware"] = tz_aware or False
        self.__options = ClientOptions(options, wconcern)
        super(MontyClient, self).__init__(self.__options.codec_options,
                                          self.__options.write_concern)

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.address == other.address
        return NotImplemented

    def __ne__(self, other):
        return not self == other

    def __repr__(self):
        return ("MontyClient({})".format(
            ", ".join([
                "repository={!r}".format(
                    self.address
                ),
                "document_class={}.{}".format(
                    self.__options._options["document_class"].__module__,
                    self.__options._options["document_class"].__name__
                ),
                "storage_engine={}".format(
                    self._storage
                ),
            ]))
        )

    def __getattr__(self, name):
        if name.startswith('_'):
            raise AttributeError(
                "MontyClient has no attribute {0!r}. To access the {0}"
                " database, use client[{0!r}].".format(name))
        return self.get_database(name)

    def __getitem__(self, key):
        return self.get_database(key)

    def __enter__(self):
        return self

    def __exit__(self, *args):
        if self._storage.is_open:
            self.close()

    @property
    def address(self):
        return self._storage.repository

    def close(self):
        self._storage.close()

    def database_names(self):
        """
        Return a list of database names.
        """
        return self._storage.database_list()

    def drop_database(self, name_or_database):
        """
        Remove database.
        # Could raise OSError: Device or resource busy
        if db file is locked by other connection...
        """
        name = name_or_database
        if isinstance(name_or_database, MontyDatabase):
            name = name_or_database.name
        elif not isinstance(name_or_database, string_type):
            raise TypeError("name_or_database must be an instance of "
                            "basestring or a Database")

        self._storage.database_drop(name)

    def get_database(self, name):
        """
        Get a database, create one if not exists.
        """
        # verify database name
        if platform.system() == "Windows":
            is_invaild = set('/\. "$*<>:|?').intersection(set(name))
        else:
            is_invaild = set('/\. "$').intersection(set(name))

        if is_invaild or not name:
            raise errors.OperationFailure("Invaild database name.")
        else:
            return MontyDatabase(self, name)

    def server_info(self):
        return {
            "version": version.version,
            "versionArray": list(version.version_info),
            "mongoVersion": version.mongo_version,
            "mongoVersionArray": list(version.mongo_version_info),
            "storageEngine": repr(self._storage),
            "python": sys.version,
            "platform": platform.platform(),
            "machine": platform.machine(),
        }





class _NoVal(object):
    def __repr__(self):
        return "_NoVal()"
    __solts__ = ()


# Value of missing field
_no_val = _NoVal()


class FieldWalkError(Exception):
    """Base class for FieldWalker exceptions."""


class FieldWriteError(FieldWalkError):
    """FieldWalker write operation error class"""

    def __init__(self, error, code=None):
        self.__code = code
        super(FieldWriteError, self).__init__(error)

    @property
    def code(self):
        return self.__code


class FieldValues(object):
    """Document field status and values iterator

    Store document tree nodes that matched from query, and iterate nodes values
    by interests. Also compute document field status from input nodes.

    Arguments:
        nodes (list): A list of nodes that picked from `FieldTree`
        fieldwalker: Instance of `FieldWalker` which made the query

    """

    __slots__ = ("nodes", "_is_exists", "_null_or_missing",
                 "_fieldwalker", "_value_iter", "__iter")

    def __init__(self, nodes, fieldwalker):
        self.nodes = nodes
        self._fieldwalker = fieldwalker

        self._is_exists = None
        self._null_or_missing = None

        self._value_iter = self.iter_full

    def is_exists(self):
        """Is field exists ?"""
        if self._is_exists is None:
            is_exists = any(nd.exists for nd in self.nodes)
            # Cache
            self._is_exists = is_exists

            return is_exists
        else:
            return self._is_exists

    def null_or_missing(self):
        """Is None value or missing field ?"""
        if self._null_or_missing is None:
            null_or_missing = (any(nd.is_missing() for nd in self.nodes) or
                               self.is_exists() and None in self.iter_flat())
            # Cache
            self._null_or_missing = null_or_missing

            return null_or_missing
        else:
            return self._null_or_missing

    def _iter(self, array_only, unpack, pack):
        """Value iterator

        Args:
            array_only (bool): Yield only `list` type values.
            unpack (bool): If value is a `list`, yield it's elements.
            pack (bool): If value is a `list`, yield it.

        """
        fieldwalker = self._fieldwalker
        for node in self.nodes:
            fieldwalker._put_matched(node)

            value = node.value
            if isinstance(value, list):
                # value in array
                if unpack and not node.located:
                    for i, elem in enumerate(value):
                        if elem is not _no_val:
                            matched = FieldNode(str(i),
                                                elem,
                                                exists=True,
                                                in_array=True,
                                                parent=node)
                            fieldwalker._put_matched(matched)
                            yield elem
                if pack:
                    # Include whole array as part of query result
                    yield value
            else:
                # Value of document field or value of positioned array element
                if not array_only and value is not _no_val:
                    yield value

        # Reset to `None` if the iter loop did not *break* in query
        fieldwalker._put_matched(None)

    def iter_plain(self):
        """Iterate each nodes value as is
        """
        return self._iter(array_only=False, unpack=False, pack=True)

    def iter_flat(self):
        """Iterate each nodes value and unpack list value's element
        """
        return self._iter(array_only=False, unpack=True, pack=False)

    def iter_full(self):
        """Iterate each nodes value as is, also unpack list value's element
        """
        return self._iter(array_only=False, unpack=True, pack=True)

    def iter_arrays(self):
        """Only iterate list type values
        """
        return self._iter(array_only=True, unpack=False, pack=True)

    def iter_elements(self):
        """Only iterate list type values' elements
        """
        return self._iter(array_only=True, unpack=True, pack=False)

    def change_iter(self, func):
        """Change value iterator

        func: A function that retures a iterator

        """
        self._value_iter = func

    def __next__(self):
        return next(self.__iter)

    next = __next__

    def __iter__(self):
        self.__iter = self._value_iter()
        return self

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self._value_iter = self.iter_full

    def __repr__(self):
        return "FieldValues({})".format(self.iter_plain())

    def __eq__(self, other):
        return list(self.iter_plain()) == other


class FieldNode(str):
    """Docuemtn field node

    Arguments:
        field (str): Document field name
        doc: field value
        located (bool): Is this a positioned field (array element)
        exists (bool): Is this field exists
        in_array (bool): Is this document inside an array
        parent (FieldNode): Parent node

    """

    # (NOTE) `__slots__` not supported for `str` in Python 2.7,
    #        drop it for now.
    #
    # __slots__ = ("value", "located", "exists", "full_path",
    #              "in_array", "parent", "children")

    def __new__(cls, field, doc, located=False, exists=False,
                in_array=False, parent=None):
        obj = str.__new__(cls, field)
        obj.value = doc
        obj.located = located
        obj.exists = exists
        obj.in_array = in_array
        obj.parent = parent
        obj.children = []

        if getattr(field, "in_array", False):
            obj.in_array = True

        return obj

    def __init__(self, *args, **kwargs):
        self.full_path = self.concat_parents()

    def __repr__(self):
        return "FieldNode({})".format(self)

    def __iter__(self):
        return iter(self.children)

    def __contains__(self, field):
        return field in self.children

    def __len__(self):
        return len(self.children)

    def __getitem__(self, field):
        for node in self.children:
            if node == field:
                return node
        raise KeyError

    def is_missing(self):
        if self.in_array and not self.located:
            return not self.exists  # doc in array, missing if not exists
        if not self.in_array and not self.exists:
            return True
        return False

    def concat_parents(self):
        forepath = getattr(self.parent, "concat_parents", lambda: "")()
        if forepath:
            return forepath + "." + str(self)
        return str(self)

    def spawn(self, value, field, located=False, exists=True, in_array=False):
        new_node = FieldNode(field, value, located, exists, in_array, self)
        self.children.append(new_node)
        return new_node


class FieldTreeReader(object):

    def __init__(self, tree):
        self.map_cls = tree.map_cls
        self.trace = set()

    def operate(self, node, field):
        if node.exists is False:
            return [node]

        result = list()

        if isinstance(node.value, self.map_cls):
            result.append(self.read_map(node, field))
        elif isinstance(node.value, list):
            result += self.read_array(node, field)
        else:
            new_node = node.spawn(_no_val,
                                  field,
                                  exists=False,
                                  located=node.located,
                                  in_array=node.in_array)
            result.append(new_node)

        return result

    def read_map(self, node, field, index=None, elem=None):
        doc = node.value if elem is None else elem

        try:
            val = doc[field]
            exists = True
        except KeyError:
            val = _no_val
            exists = False

        if index:
            field = index + "." + field

        self.trace.add(field)
        return node.spawn(val, field, exists=exists, in_array=bool(index))

    def read_array(self, node, field):
        new_nodes = list()
        doc = node.value

        for i, elem in enumerate(doc):
            if isinstance(elem, self.map_cls):
                new_nodes.append(self.read_map(node, field, str(i), elem))

        if field.isdigit():
            try:
                val = doc[int(field)]
                exists = True
            except IndexError:
                val = _no_val
                exists = False

            self.trace.add(field)
            new_nodes.append(node.spawn(val,
                                        field,
                                        located=True,
                                        exists=exists,
                                        in_array=True))
        return new_nodes


def is_multi_position_operator(field):
    return field[:2] == "$["


def parse_identifier(field):
    return field[2:-1]  # $[identifier]


_dollar_prefixed_err_msg = (
    "The dollar ($) prefixed field {0!r} in {1!r} is not valid for storage."
)


def no_dollar_prefix_field(doc, map_cls, root_path):
    if isinstance(doc, map_cls):
        for field, value in doc.items():
            if field.startswith("$"):
                full_path = root_path + "." + field
                msg = _dollar_prefixed_err_msg.format(field, full_path)
                raise FieldWriteError(msg, code=52)

            no_dollar_prefix_field(value, map_cls, root_path + "." + field)

    if isinstance(doc, list):
        for i, elem in enumerate(doc):
            no_dollar_prefix_field(elem, map_cls, root_path + "." + str(i))


class FieldTreeWriter(object):

    def __init__(self, tree):
        self.map_cls = tree.map_cls
        self.filters = None
        self.on_delete = False
        self.trace = set()

    def operate(self, node, field):
        result = list()

        if not field:
            raise FieldWriteError("An empty update path is not valid", code=56)

        if field.startswith("."):
            msg = ("The update path {} contains an empty field name, which is "
                   "not allowed".format(field))
            raise FieldWriteError(msg, code=56)

        if field.startswith("$") and not field.startswith("$["):
            full_path = node.full_path + "." + field
            msg = _dollar_prefixed_err_msg.format(field, full_path)
            raise FieldWriteError(msg, code=52)

        if not node.exists and is_multi_position_operator(field):
            msg = ("The path {0!r} must exist in the document in order "
                   "to apply array updates.".format(node.full_path))
            raise FieldWriteError(msg, code=2)

        if (is_multi_position_operator(field) and
                not isinstance(node.value, list)):
            msg = ("Cannot apply array updates to non-array "
                   "element {0}: {1}".format(str(node), node.value))
            raise FieldWriteError(msg, code=2)

        if isinstance(node.value, self.map_cls):

            result.append(self.write_map(node, field))

        elif (isinstance(node.value, list) and
                (field.isdigit() or is_multi_position_operator(field))):

            result += self.write_array(node, field)

        elif not self.on_delete:
            msg = ("Cannot create field {0!r} in element "
                   "{1}".format(field, {str(node): node.value}))
            raise FieldWriteError(msg, code=28)

        return result

    def write_map(self, node, field, index=None, elem=None):
        doc = node.value if elem is None else elem

        try:
            val = doc[field]
            exists = True
        except KeyError:
            val = self.map_cls()
            exists = False

        if index:
            field = index + "." + field

        self.trace.add(field)
        return node.spawn(val, field, exists=exists, in_array=bool(index))

    def write_array(self, node, field):
        new_nodes = list()
        doc = node.value

        if is_multi_position_operator(field):
            identifier = parse_identifier(field)
            if identifier:
                # filter
                filter = self.filters[identifier]
                for i, elem in enumerate(doc):
                    if filter({identifier: elem}):
                        field = str(i)

                        self.trace.add(field)
                        new_nodes.append(node.spawn(elem,
                                                    field,
                                                    located=True,
                                                    exists=True,
                                                    in_array=True))
            else:
                # all
                for i, elem in enumerate(doc):
                    field = str(i)

                    self.trace.add(field)
                    new_nodes.append(node.spawn(elem,
                                                field,
                                                located=True,
                                                exists=True,
                                                in_array=True))
        else:
            try:
                val = doc[int(field)]
                exists = True
            except IndexError:
                val = self.map_cls()
                exists = False

            self.trace.add(field)
            new_nodes.append(node.spawn(val,
                                        field,
                                        located=True,
                                        exists=exists,
                                        in_array=True))
        return new_nodes


def is_conflict(path_a, path_b):
    return (path_a == path_b or
            path_a.startswith(path_b + ".") or
            path_b.startswith(path_a + "."))


class FieldTree(object):

    def __init__(self, doc, doc_type=None):
        self.map_cls = doc_type or type(doc)
        self.root = FieldNode("", doc, exists=True)
        self.clear()

    def __str__(self):
        def print_tree(node, level=0):
            status = "*" * ((not node.exists) + node.is_missing())
            tree_str = "\t" * level + node + status + "{}\n"

            if len(node) == 0:
                tree_str = tree_str.format(": " + str(node.value))
            else:
                tree_str = tree_str.format("")

            for child in node:
                tree_str += print_tree(child, level + 1)

            return tree_str

        return "FieldTree({})".format(print_tree(self.root))

    def __repr__(self):
        return "FieldTree({})".format(self)

    def clear(self):
        self.root.children = list()
        self.handler = None
        self.changes = list()
        self.restart()

    def restart(self):
        self.picked = [self.root]
        self.previous = set([""])

    def grow(self, fields):

        for field in fields:

            if not is_multi_position_operator(field):
                # Reuse previous spawned nodes
                old_picked = []
                for node in [child for node in self.picked for child in node]:
                    if field == node and node.parent in self.previous:
                        old_picked.append(node)

                if old_picked:
                    self.picked = old_picked
                    self.previous.add(field)
                    continue

            self.handler.trace.clear()
            new_picked = []
            for node in self.picked:
                if node not in self.previous:
                    continue

                new_picked += self.handler.operate(node, field)

            self.picked = new_picked

            self.previous = self.handler.trace.copy()

            # When READ, stop if all nodes not exists
            if (isinstance(self.handler, FieldTreeReader) and
                    all(node.exists is False for node in self.picked)):
                break

    def read(self, fields):
        self.handler = FieldTreeReader(self)
        self.grow(fields)
        return self.picked

    def stage(self, value, evaluator=None):
        """Internal method, for staging the changes
        """
        evaluator = evaluator or (lambda _, val: val)
        staging = [node for node in self.picked if node in self.previous]
        updates = list()

        for node in staging:
            update = node.full_path
            for changed in self.changes:
                if is_conflict(update, changed):
                    msg = ("Updating the path {0!r} would create a conflict "
                           "at {1!r}".format(update, changed))
                    raise FieldWriteError(msg, code=40)

            new_value = evaluator(node, value)

            if node.value != new_value:
                no_dollar_prefix_field(new_value, self.map_cls, node.full_path)
                node.value = new_value
                updates.append(update)

        self.changes += updates

    def fields_positioning(self, fieldwalker, array_filters=None):
        fields = fieldwalker.steps

        if "$" in fields:

            if not fieldwalker.has_matched():
                # If hasn't queried or not matched in array
                msg = ("The positional operator did not find the match needed "
                       "from the query.")
                raise FieldWriteError(msg, code=2)

            elif fields.count("$") > 1:
                msg = ("Too many positional (i.e. '$') elements found in path "
                       "{!r}".format(".".join(fields)))
                raise FieldWriteError(msg, code=2)

            else:
                # Replace "$" into top matched array element index
                position_path = ".".join(fields).split(".$", 1)[0]
                top_matched = fieldwalker.top_matched(position_path)
                position = top_matched.split(".")[0]
                fields[fields.index("$")] = position

        if array_filters is None:
            return fields

        for field in fields:
            if is_multi_position_operator(field):
                identifier = parse_identifier(field)

                if identifier and identifier not in array_filters:
                    msg = ("No array filter found for identifier {0!r} in "
                           "path {1!r}".format(identifier, ".".join(fields)))
                    raise FieldWriteError(msg, code=2)
        return fields

    def write(self, fields, value, evaluator=None, array_filters=None):
        self.handler = FieldTreeWriter(self)
        self.handler.filters = array_filters
        self.handler.on_delete = value is _no_val
        self.grow(fields)
        self.stage(value, evaluator=evaluator)

    def delete(self, fields, array_filters=None):
        self.write(fields, _no_val, array_filters=array_filters)

    def extract(self, visited_only=False):
        ON_DELETE = False
        if isinstance(self.handler, FieldTreeWriter):
            ON_DELETE = self.handler.on_delete

        def _extract(node, visited_only):
            doc = node.value
            has_children = bool(node.children)
            visited_only_ = visited_only

            if isinstance(doc, self.map_cls):
                new_doc = self.map_cls()
                fields = list(doc.keys()) + [str(child) for child in node
                                             if child not in doc]
                for field in fields:
                    try:
                        child = node[field]
                    except KeyError:
                        if visited_only and has_children:
                            visited_only_ = False
                            continue
                        value = doc[field]
                    else:
                        value = _extract(child, visited_only_)

                    if value is not _no_val:
                        new_doc[field] = value

                return new_doc

            elif isinstance(doc, list):
                new_doc = list()
                indices = range(max(
                    [len(doc)] + [int(n) + 1 for n in node if n.isdigit()]))

                for index in indices:
                    if ON_DELETE and index >= len(doc):
                        continue

                    try:
                        child = node[str(index)]
                    except KeyError:
                        if visited_only and has_children:
                            visited_only_ = False
                            continue
                        value = doc[index] if index < len(doc) else None
                    else:
                        value = _extract(child, visited_only_)

                    if value is _no_val:
                        value = None
                    new_doc.append(value)

                return new_doc

            else:
                return doc

        return _extract(self.root, visited_only)


class FieldWalker(object):
    """Document field traversal interface for MontyDB

    Key component to make everything run. Every document operation in MontyDB
    use this to read/write the field value into document or to delete field.

    The `FieldWalker` does not interact with the input document directly, it
    initialized with a tree (`FieldTree`), and adding nodes (`FieldNode`) via
    the input field.

    when read, all matched fields' values will be carry out by a `FieldValues`
    instance.

    When write/delete, you need to commit the changes before it take effect.

    Example:
        >>> from montydb.engine.core import FieldWalker
        >>> doc = {"name": "Tom", "detail": {"age": None, "weight": 80}}
        >>> walker = FieldWalker(doc)
        >>> walker.go("name").get().value
        FieldValues(['Tom'])
        >>> walker.go("detail.weight").get().value
        FieldValues([80])
        >>> walker.go("detail.age").set(30)
        >>> walker.commit()
        True
        >>> walker.doc
        {'name': 'Tom', 'detail': {'age': 30, 'weight': 80}}

    Arguments:
        doc: The document to operate on. Must be a mutable mapping type.
        doc_type (optional): The document class, specify what type of the
            document is. If `doc_type` not provided, will use `type(doc)`
            to get document class.

    """

    def __init__(self, doc, doc_type=None):
        self.doc = doc
        self.doc_type = doc_type or type(doc)
        self.steps = None
        self.tree = FieldTree(doc, doc_type)
        self.value = None
        self.path = None
        self.matched = dict()

    def go(self, path):
        """Input document traversing field path

        https://docs.mongodb.com/manual/core/document/#dot-notation

        Arguments:
            path (str): Document field path

        Returns:
            (FieldWalker): self

        """
        self.tree.restart()
        self.path = path
        self.steps = path.split(".")
        return self

    def step(self, field):
        """Push traversing one depth further

        Arguments:
            field (str): Document field name

        Returns:
            (FieldWalker): self

        """
        if self.path is None:
            self.path = field
        else:
            self.path += "." + field
        self.steps = [field]
        return self

    def restart(self):
        """Reset traversing state

        Returns:
            (FieldWalker): self

        """
        self.tree.restart()
        self.path = None
        return self

    def get(self):
        """Read all values from previous given field path

        Returns:
            (FieldWalker): self

        """
        self.value = FieldValues(self.tree.read(self.steps), self)
        return self

    def set(self, value, evaluator=None, array_filters=None):
        """Write value into fields that matched

        This operation will not take effect util `commit()`.

        Arguments:
            value: Any value that needs to be written into
            evaluator (optional): A function which compute the result from
                `value` and the current field value from document, and the
                result will be the new value to write into document.
                This function should take two arguments, the first arg is an
                instance of `.engine.core.field_walker.FieldNode`, the second
                is the `value`.
            array_filters (optional): An array filter document which formed as
                `{<top-field-name>: <.engine.queries.QueryFilter>, ..}`.

        """
        steps = self.tree.fields_positioning(self, array_filters)
        self.tree.write(steps, value, evaluator, array_filters)

    def drop(self, array_filters=None):
        """Delete field

        This operation will not take effect util `commit()`.

        Arguments:
            array_filters (optional): An array filter document which formed as
                `{<top-field-name>: <.engine.queries.QueryFilter>, ..}`.

        """
        steps = self.tree.fields_positioning(self, array_filters)
        self.tree.delete(steps, array_filters)

    def commit(self):
        """Update document if any change been staged

        Returns:
            bool: Return `True` if any change happened, else `False`

        """
        has_change = bool(self.tree.changes)
        if has_change:
            self.doc = self.tree.extract()
        return has_change

    def touched(self):
        """Return a copy of document that contains only visited fields

        Returns:
            document

        """
        return self.tree.extract(visited_only=True)

    def top_matched(self, position_path):
        for path, node in self.matched.items():
            if path.startswith(position_path + "."):
                matched = node
                break
        else:
            return

        first = None
        while matched.parent is not None:
            if not matched.in_array:
                break
            first = matched
            matched = matched.parent

        return first

    def has_matched(self):
        return any(node.in_array for node in self.matched.values())

    def _put_matched(self, node):
        if node is None:
            self.matched.pop(self.path, None)
        else:
            self.matched[self.path] = node

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.tree.clear()



from .base import (
    ASCENDING,
    DESCENDING,
)
from .client import MontyClient
from .database import MontyDatabase
from .collection import MontyCollection
from .cursor import (
    MontyCursor,
    CursorType
)
from .configure import (
    pin_repo,
    current_repo,
    open_repo,
    set_storage,
)

from . import utils

from .version import (
    version,
    version_info,
    mongo_version,
    mongo_version_info,
    __version__,
)


__all__ = [
    "MontyClient",
    "MontyDatabase",
    "MontyCollection",
    "MontyCursor",

    "pin_repo",
    "current_repo",
    "open_repo",
    "set_storage",

    "ASCENDING",
    "DESCENDING",
    "CursorType",

    "utils",

    "version",
    "version_info",
    "mongo_version",
    "mongo_version_info",
    "__version__",
]



# Copyright 2011-present MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# Modifications copyright (C) 2018 davidlatwe
#
# Assembling crucial classes and functions form pymongo module,
# some of them may modified by needs.


import collections

from bson import BSON, SON
from bson.py3compat import integer_types, string_type, abc, iteritems
from bson.codec_options import CodecOptions
from bson.raw_bson import RawBSONDocument
from bson.codec_options import _parse_codec_options

from .engine.helpers import is_duckument_type

ASCENDING = 1
DESCENDING = -1


def command_coder(*args, **kwargs):
    """Convert command doc to SON type
    """
    codec_op = kwargs["codec_op"]
    for doc in args:
        if is_duckument_type(doc):
            # Preserve the type of the query document when decode it back
            order_keep = codec_op.with_options(document_class=SON)
            yield BSON.encode(doc, False, codec_op).decode(order_keep)
        else:
            yield doc  # should be None type.


def validate_is_document_type(option, value):
    """Validate the type of method arguments that expect a MongoDB document."""
    if not isinstance(value, (collections.MutableMapping, RawBSONDocument)):
        raise TypeError("%s must be an instance of dict, bson.son.SON, "
                        "bson.raw_bson.RawBSONDocument, or "
                        "a type that inherits from "
                        "collections.MutableMapping" % (option,))


def validate_boolean(option, value):
    """Validates that 'value' is True or False."""
    if isinstance(value, bool):
        return value
    raise TypeError("%s must be True or False" % (option,))


def validate_list(option, value):
    """Validates that 'value' is a list."""
    if not isinstance(value, list):
        raise TypeError("%s must be a list" % (option,))
    return value


def validate_list_or_none(option, value):
    """Validates that 'value' is a list or None."""
    if value is None:
        return value
    return validate_list(option, value)


def validate_is_mapping(option, value):
    """Validate the type of method arguments that expect a document."""
    if not isinstance(value, abc.Mapping):
        raise TypeError("%s must be an instance of dict, bson.son.SON, or "
                        "other type that inherits from "
                        "collections.Mapping" % (option,))


def validate_ok_for_update(update):
    """Validate an update document."""
    validate_is_mapping("update", update)
    # Update can not be {}
    if not update:
        raise ValueError('update only works with $ operators')
    first = next(iter(update))
    if not first.startswith('$'):
        raise ValueError('update only works with $ operators')


def validate_ok_for_replace(replacement):
    """Validate a replacement document."""
    validate_is_mapping("replacement", replacement)
    # Replacement can be {}
    if replacement and not isinstance(replacement, RawBSONDocument):
        first = next(iter(replacement))
        if first.startswith('$'):
            raise ValueError('replacement can not include $ operators')


def _fields_list_to_dict(fields, option_name):
    """Takes a sequence of field names and returns a matching dictionary.
    ["a", "b"] becomes {"a": 1, "b": 1}
    and
    ["a.b.c", "d", "a.c"] becomes {"a.b.c": 1, "d": 1, "a.c": 1}
    """
    if isinstance(fields, abc.Mapping):
        return fields

    if isinstance(fields, (abc.Sequence, abc.Set)):
        if not all(isinstance(field, string_type) for field in fields):
            raise TypeError("%s must be a list of key names, each an "
                            "instance of %s" % (option_name,
                                                string_type.__name__))
        return dict.fromkeys(fields, 1)

    raise TypeError("%s must be a mapping or "
                    "list of key names" % (option_name,))


def _index_list(key_or_list, direction=None):
    """Helper to generate a list of (key, direction) pairs.

    Takes such a list, or a single key, or a single key and direction.
    """
    if direction is not None:
        return [(key_or_list, direction)]
    else:
        if isinstance(key_or_list, string_type):
            return [(key_or_list, ASCENDING)]
        elif not isinstance(key_or_list, (list, tuple)):
            raise TypeError("if no direction is specified, "
                            "key_or_list must be an instance of list")
        return key_or_list


def _index_document(index_list):
    """Helper to generate an index specifying document.

    Takes a list of (key, direction) pairs.
    """
    if isinstance(index_list, abc.Mapping):
        raise TypeError("passing a dict to sort/create_index/hint is not "
                        "allowed - use a list of tuples instead. did you "
                        "mean %r?" % list(iteritems(index_list)))
    elif not isinstance(index_list, (list, tuple)):
        raise TypeError("must use a list of (key, direction) pairs, "
                        "not: " + repr(index_list))
    if not len(index_list):
        raise ValueError("key_or_list must not be the empty list")

    index = SON()
    for (key, value) in index_list:
        if not isinstance(key, string_type):
            raise TypeError("first item in each key pair must be a string")
        if not isinstance(value, (string_type, int, abc.Mapping)):
            raise TypeError("second item in each key pair must be 1, -1, "
                            "'2d', 'geoHaystack', or another valid MongoDB "
                            "index specifier.")
        index[key] = value
    return index


class WriteConcern(object):
    """MontyWriteConcern
    """

    __slots__ = ("_document")

    def __init__(self, wtimeout=None):
        self._document = {}

        if wtimeout is not None:
            if not isinstance(wtimeout, integer_types):
                raise TypeError("wtimeout must be an integer")
            self._document["wtimeout"] = wtimeout

    @property
    def document(self):
        """The document representation of this write concern.
        """
        return self._document.copy()

    def __repr__(self):
        return ("MontyWriteConcern({})".format(
            ", ".join("%s=%s" % kvt for kvt in self.document.items()),))

    def __eq__(self, other):
        return self.document == other.document

    def __ne__(self, other):
        return self.document != other.document

    def __bool__(self):
        return bool(self.document)


def _parse_write_concern(options):
    """Parse write concern options."""
    wtimeout = options.get('wtimeout')
    return WriteConcern(wtimeout)


class ClientOptions(object):
    """ClientOptions"""

    def __init__(self, options, storage_wconcern=None):
        self.__options = options
        self.__codec_options = _parse_codec_options(options)

        if storage_wconcern is not None:
            self.__write_concern = storage_wconcern
        else:
            self.__write_concern = _parse_write_concern(options)

    @property
    def _options(self):
        """The original options used to create this ClientOptions."""
        return self.__options  # pragma: no cover

    @property
    def codec_options(self):
        """A :class:`~bson.codec_options.CodecOptions` instance."""
        return self.__codec_options

    @property
    def write_concern(self):
        """A :class:`~montydb.base.WriteConcern` instance."""
        return self.__write_concern


class BaseObject(object):
    """A base class that provides attributes and methods common
    to multiple montydb classes.

    SHOULD NOT BE USED BY DEVELOPERS EXTERNAL TO MONTYDB.
    """

    def __init__(self, codec_options, write_concern):

        if not isinstance(codec_options, CodecOptions):
            raise TypeError("codec_options must be an "  # pragma: no cover
                            "instance of bson.codec_options.CodecOptions")
        self.__codec_options = codec_options

        if not isinstance(write_concern, WriteConcern):
            raise TypeError("write_concern must be an "  # pragma: no cover
                            "instance of montydb.base.WriteConcern")
        self.__write_concern = write_concern

    @property
    def codec_options(self):
        """Read only access to the :class:`~bson.codec_options.CodecOptions`
        of this instance.
        """
        return self.__codec_options

    @property
    def write_concern(self):
        """Read only access to the :class:`~montydb.base.WriteConcern`
        of this instance.
        """
        return self.__write_concern
